#!/usr/bin/env python

"""
Interpretation and preparation of iMIP content, together with a content handling
mechanism employed by specific recipients.

Copyright (C) 2014, 2015 Paul Boddie <paul@boddie.org.uk>

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 3 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
details.

You should have received a copy of the GNU General Public License along with
this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from datetime import datetime, timedelta
from email.mime.text import MIMEText
from imiptools.dates import *
from imiptools.period import have_conflict, insert_period, remove_period
from pytz import timezone
from vCalendar import parse, ParseError, to_dict
from vRecurrence import get_parameters, get_rule
import email.utils
import imip_store

try:
    from cStringIO import StringIO
except ImportError:
    from StringIO import StringIO

# Content interpretation.

def get_items(d, name, all=True):

    """
    Get all items from 'd' with the given 'name', returning single items if
    'all' is specified and set to a false value and if only one value is
    present for the name. Return None if no items are found for the name.
    """

    if d.has_key(name):
        values = d[name]
        if not all and len(values) == 1:
            return values[0]
        else:
            return values
    else:
        return None

def get_item(d, name):
    return get_items(d, name, False)

def get_value_map(d, name):

    """
    Return a dictionary for all items in 'd' having the given 'name'. The
    dictionary will map values for the name to any attributes or qualifiers
    that may have been present.
    """

    items = get_items(d, name)
    if items:
        return dict(items)
    else:
        return {}

def get_values(d, name, all=True):
    if d.has_key(name):
        values = d[name]
        if not all and len(values) == 1:
            return values[0][0]
        else:
            return map(lambda x: x[0], values)
    else:
        return None

def get_value(d, name):
    return get_values(d, name, False)

def get_utc_datetime(d, name):
    value, attr = get_item(d, name)
    dt = get_datetime(value, attr)
    return to_utc_datetime(dt)

def get_addresses(values):
    return [address for name, address in email.utils.getaddresses(values)]

def get_address(value):
    return value.lower().startswith("mailto:") and value.lower()[7:] or value

def get_uri(value):
    return value.lower().startswith("mailto:") and value.lower() or ":" in value and value or "mailto:%s" % value.lower()

def uri_dict(d):
    return dict([(get_uri(key), value) for key, value in d.items()])

def uri_item(item):
    return get_uri(item[0]), item[1]

def uri_items(items):
    return [(get_uri(value), attr) for value, attr in items]

def is_new_object(old_sequence, new_sequence, old_dtstamp, new_dtstamp, partstat_set):

    """
    Return for the given 'old_sequence' and 'new_sequence', 'old_dtstamp' and
    'new_dtstamp', and the 'partstat_set' indication, whether the object
    providing the new information is really newer than the object providing the
    old information.
    """

    have_sequence = old_sequence is not None and new_sequence is not None
    is_same_sequence = have_sequence and int(new_sequence) == int(old_sequence)

    have_dtstamp = old_dtstamp and new_dtstamp
    is_old_dtstamp = have_dtstamp and new_dtstamp < old_dtstamp or old_dtstamp and not new_dtstamp

    is_old_sequence = have_sequence and (
        int(new_sequence) < int(old_sequence) or
        is_same_sequence and is_old_dtstamp
        )

    return is_same_sequence and partstat_set or not is_old_sequence

# NOTE: Need to expose the 100 day window for recurring events in the
# NOTE: configuration.

def get_periods(obj, window_size=100):

    """
    Return periods for the given object 'obj', confining materialised periods
    to the given 'window_size' in days starting from the present moment.
    """

    dtstart = get_utc_datetime(obj, "DTSTART")
    dtend = get_utc_datetime(obj, "DTEND")

    # NOTE: Need also DURATION support.

    duration = dtend - dtstart

    # Recurrence rules create multiple instances to be checked.
    # Conflicts may only be assessed within a period defined by policy
    # for the agent, with instances outside that period being considered
    # unchecked.

    window_end = datetime.now() + timedelta(window_size)

    # NOTE: Need also RDATE and EXDATE support.

    rrule = get_value(obj, "RRULE")

    if rrule:
        selector = get_rule(dtstart, rrule)
        parameters = get_parameters(rrule)
        periods = []
        for start in selector.materialise(dtstart, window_end, parameters.get("COUNT"), parameters.get("BYSETPOS")):
            start = datetime(*start, tzinfo=timezone("UTC"))
            end = start + duration
            periods.append((format_datetime(start), format_datetime(end)))
    else:
        periods = [(format_datetime(dtstart), format_datetime(dtend))]

    return periods

def remove_from_freebusy(freebusy, attendee, uid, store):

    """
    For the given 'attendee', remove periods from 'freebusy' that are associated
    with 'uid' in the 'store'.
    """

    remove_period(freebusy, uid)
    store.set_freebusy(attendee, freebusy)

def remove_from_freebusy_for_other(freebusy, user, other, uid, store):

    """
    For the given 'user', remove for the 'other' party periods from 'freebusy'
    that are associated with 'uid' in the 'store'.
    """

    remove_period(freebusy, uid)
    store.set_freebusy_for_other(user, freebusy, other)

def _update_freebusy(freebusy, periods, transp, uid):

    """
    Update the free/busy details with the given 'periods', 'transp' setting and
    'uid'.
    """

    remove_period(freebusy, uid)

    for start, end in periods:
        insert_period(freebusy, (start, end, uid, transp))

def update_freebusy(freebusy, attendee, periods, transp, uid, store):

    """
    For the given 'attendee', update the free/busy details with the given
    'periods', 'transp' setting and 'uid' in the 'store'.
    """

    _update_freebusy(freebusy, periods, transp, uid)
    store.set_freebusy(attendee, freebusy)

def update_freebusy_for_other(freebusy, user, other, periods, transp, uid, store):

    """
    For the given 'user', update the free/busy details of 'other' with the given
    'periods', 'transp' setting and 'uid' in the 'store'.
    """

    _update_freebusy(freebusy, periods, transp, uid)
    store.set_freebusy_for_other(user, freebusy, other)

def can_schedule(freebusy, periods, uid):

    """
    Return whether the 'freebusy' list can accommodate the given 'periods'
    employing the specified 'uid'.
    """

    for conflict in have_conflict(freebusy, periods, True):
        start, end, found_uid, found_transp = conflict
        if found_uid != uid:
            return False

    return True

# Handler mechanism objects.

def handle_itip_part(part, senders, recipients, handlers, messenger):

    """
    Handle the given iTIP 'part' from the given 'senders' for the given
    'recipients' using the given 'handlers' and information provided by the
    given 'messenger'. Return a list of responses, each response being a tuple
    of the form (is-outgoing, message-part).
    """

    method = part.get_param("method")

    # Decode the data and parse it.

    f = StringIO(part.get_payload(decode=True))

    itip = parse_object(f, part.get_content_charset(), "VCALENDAR")

    # Ignore the part if not a calendar object.

    if not itip:
        return []

    # Require consistency between declared and employed methods.

    if get_value(itip, "METHOD") == method:

        # Look for different kinds of sections.

        all_results = []

        for name, cls in handlers:
            for details in get_values(itip, name) or []:

                # Dispatch to a handler and obtain any response.

                handler = cls(details, senders, recipients, messenger)
                result = methods[method](handler)()

                # Aggregate responses for a single message.

                if result:
                    response_method, part = result
                    outgoing = method != response_method
                    all_results.append((outgoing, part))

        return all_results

    return []

def parse_object(f, encoding, objtype=None):

    """
    Parse the iTIP content from 'f' having the given 'encoding'. If 'objtype' is
    given, only objects of that type will be returned. Otherwise, the root of
    the content will be returned as a dictionary with a single key indicating
    the object type.

    Return None if the content was not readable or suitable.
    """

    try:
        try:
            doctype, attrs, elements = obj = parse(f, encoding=encoding)
            if objtype and doctype == objtype:
                return to_dict(obj)[objtype][0]
            elif not objtype:
                return to_dict(obj)
        finally:
            f.close()

    # NOTE: Handle parse errors properly.

    except (ParseError, ValueError):
        pass

    return None

def to_part(method, calendar):

    """
    Write using the given 'method', the 'calendar' details to a MIME
    text/calendar part.
    """

    encoding = "utf-8"
    out = StringIO()
    try:
        imip_store.to_stream(out, imip_store.make_calendar(calendar, method), encoding)
        part = MIMEText(out.getvalue(), "calendar", encoding)
        part.set_param("method", method)
        return part

    finally:
        out.close()

class Handler:

    "General handler support."

    def __init__(self, details, senders=None, recipients=None, messenger=None):

        """
        Initialise the handler with the 'details' of a calendar object and the
        'senders' and 'recipients' of the object (if specifically indicated).
        """

        self.details = details
        self.senders = senders and set(map(get_address, senders))
        self.recipients = recipients and set(map(get_address, recipients))
        self.messenger = messenger

        self.uid = get_value(details, "UID")
        self.sequence = get_value(details, "SEQUENCE")
        self.dtstamp = get_value(details, "DTSTAMP")

        self.store = imip_store.FileStore()

        try:
            self.publisher = imip_store.FilePublisher()
        except OSError:
            self.publisher = None

    # Access to calendar structures and other data.

    def get_items(self, name, all=True):
        return get_items(self.details, name, all)

    def get_item(self, name):
        return get_item(self.details, name)

    def get_value_map(self, name):
        return get_value_map(self.details, name)

    def get_values(self, name, all=True):
        return get_values(self.details, name, all)

    def get_value(self, name):
        return get_value(self.details, name)

    def get_utc_datetime(self, name):
        return get_utc_datetime(self.details, name)

    def get_periods(self):
        return get_periods(self.details)

    def remove_from_freebusy(self, freebusy, attendee):
        remove_from_freebusy(freebusy, attendee, self.uid, self.store)

    def remove_from_freebusy_for_other(self, freebusy, user, other):
        remove_from_freebusy_for_other(freebusy, user, other, self.uid, self.store)

    def update_freebusy(self, freebusy, attendee, periods):
        return update_freebusy(freebusy, attendee, periods, self.get_value("TRANSP"), self.uid, self.store)

    def update_freebusy_for_other(self, freebusy, user, other, periods):
        return update_freebusy_for_other(freebusy, user, other, periods, self.get_value("TRANSP"), self.uid, self.store)

    def can_schedule(self, freebusy, periods):
        return can_schedule(freebusy, periods, self.uid)

    def filter_by_senders(self, mapping):

        """
        Return a list of items from 'mapping' filtered using sender information.
        """

        if self.senders:

            # Get a mapping from senders to identities.

            identities = self.get_sender_identities(mapping)

            # Find the senders that are valid.

            senders = map(get_address, identities)
            valid = self.senders.intersection(senders)

            # Return the true identities.

            return [identities[get_uri(address)] for address in valid]
        else:
            return mapping

    def filter_by_recipients(self, mapping):

        """
        Return a list of items from 'mapping' filtered using recipient
        information.
        """

        if self.recipients:
            addresses = map(get_address, mapping)
            return map(get_uri, self.recipients.intersection(addresses))
        else:
            return mapping

    def require_organiser_and_attendees(self, from_organiser=True):

        """
        Return the organiser and attendees for the current object, filtered by
        the recipients of interest. Return None if no identities are eligible.

        Organiser and attendee identities are provided as lower case values.
        """

        attendee_map = uri_dict(self.get_value_map("ATTENDEE"))
        organiser_item = uri_item(self.get_item("ORGANIZER"))

        # Only provide details for attendees who sent/receive the message.

        attendee_filter_fn = from_organiser and self.filter_by_recipients or self.filter_by_senders

        attendees = {}
        for attendee in attendee_filter_fn(attendee_map):
            attendees[attendee] = attendee_map[attendee]

        if not attendees or not organiser_item:
            return None

        # Only provide details for an organiser who sent/receives the message.

        organiser_filter_fn = from_organiser and self.filter_by_senders or self.filter_by_recipients

        if not organiser_filter_fn(dict([organiser_item])):
            return None

        return organiser_item, attendees

    def get_sender_identities(self, mapping):

        """
        Return a mapping from actual senders to the identities for which they
        have provided data, extracting this information from the given
        'mapping'.
        """

        senders = {}

        for value, attr in mapping.items():
            sent_by = attr.get("SENT-BY")
            if sent_by:
                senders[get_uri(sent_by)] = value
            else:
                senders[value] = value

        return senders

    def get_object(self, user, objtype):

        """
        Return the stored object to which the current object refers for the
        given 'user' and for the given 'objtype'.
        """

        f = self.store.get_event(user, self.uid)
        obj = f and parse_object(f, "utf-8", objtype)
        return obj

    def have_new_object(self, attendee, objtype, obj=None):

        """
        Return whether the current object is new to the 'attendee' for the
        given 'objtype'.
        """

        obj = obj or self.get_object(attendee, objtype)

        # If found, compare SEQUENCE and potentially DTSTAMP.

        if obj:
            sequence = get_value(obj, "SEQUENCE")
            dtstamp = get_value(obj, "DTSTAMP")

            # NOTE: Some clients like Claws Mail erase time information from DTSTAMP
            # NOTE: and make it invalid. Thus, attendance information may also be
            # NOTE: checked.

            for _attendee, old_attr in get_items(obj, "ATTENDEE"):
                if _attendee == attendee:
                    break
            else:
                return False

            for _attendee, new_attr in self.get_items("ATTENDEE"):
                if _attendee == attendee:
                    break
            else:
                return False

            old_partstat = old_attr.get("PARTSTAT")
            new_partstat = new_attr.get("PARTSTAT")

            partstat_set = old_partstat == "NEEDS-ACTION" and new_partstat and \
                new_partstat != old_partstat

            # If the request refers to an older version of the object, ignore
            # it.

            return is_new_object(sequence, self.sequence, dtstamp, self.dtstamp, partstat_set)

        return True

    def update_dtstamp(self):

        "Update the DTSTAMP in the current object."

        dtstamp = self.get_utc_datetime("DTSTAMP")
        utcnow = to_timezone(datetime.utcnow(), "UTC")
        self.details["DTSTAMP"] = [(format_datetime(dtstamp > utcnow and dtstamp or utcnow), {})]

# Handler registry.

methods = {
    "ADD"            : lambda handler: handler.add,
    "CANCEL"         : lambda handler: handler.cancel,
    "COUNTER"        : lambda handler: handler.counter,
    "DECLINECOUNTER" : lambda handler: handler.declinecounter,
    "PUBLISH"        : lambda handler: handler.publish,
    "REFRESH"        : lambda handler: handler.refresh,
    "REPLY"          : lambda handler: handler.reply,
    "REQUEST"        : lambda handler: handler.request,
    }

# vim: tabstop=4 expandtab shiftwidth=4
